#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
import collections
import copy
import json
import os, os.path
import hashlib
from urllib.parse import urlparse, urljoin
import shutil
from tempfile import NamedTemporaryFile
from .metadata import Metadata
from .index import Index
from ..json import GenericJSONEncoder


logger = logging.getLogger('model.RepoConfig')
class RepoConfig(collections.MutableMapping):
    EMPTY = {"apps":[]}

    def __init__(self,url,cfg,config):
        self.__config = config
        self.__url = self.__clean_url(url)
        self.__store = {**RepoConfig.EMPTY,**cfg}
        self.__store['id'] = hashlib.sha1(str(self.__url).encode('UTF-8')).hexdigest()

    def __clean_url(self,url):
        p = urlparse(url)
        url = "%s://%s%s"%(p.scheme,p.netloc,p.path)
        if url.endswith('index.xml'): url = url[:-9]
        if url.endswith('index.jar'): url = url[:-9]
        if url.endswith('index-v1.jar'): url = url[:-12]
        if url.endswith('index-v1.json'): url = url[:-13]
        if not url.endswith('/'): url += '/'
        return url

    @property
    def index(self):
        return self.__config.index(self.url)

    @property
    def auth(self):
        auth = self.__store.get('auth',None)
        if auth is None or len(auth) != 2:
            return None
        return (auth[0],auth[1])

    @property
    def verify(self):
        return self.__store.get('ssl_verify',True)

    @property
    def id(self):
        return self.__store.get('id',None)

    @property
    def key(self):
        return self.url

    @property
    def url(self):
        return str(self.__url)

    @property
    def hash(self):
        return self.__store.get('hash',None)

    @property
    def filename(self):
        return os.path.join(self.__config.cache_dir,self.id+".cache")

    @property
    def format(self):
        if 'format' in self.__store:
            return self.__store['format']
        return None

    @property
    def url_index(self):
        return urljoin(self.__url,'index.jar')

    @property
    def url_index_v1(self):
        return urljoin(self.__url,'index-v1.jar')

    @property
    def src_download(self):
        if 'src_download' in self.__store: return self.__store['src_download'] == True
        return self.__config.src_download == True

    @property
    def metadata_download(self):
        if 'metadata_download' in self.__store: return self.__store['metadata_download'] == True
        return self.__config.metadata_download == True

    @property
    def default_locale(self):
        if 'default_locale' in self.__store: return str(self.__store['default_locale'])
        return self.__config.default_locale

    @property
    def versions(self):
        if 'versions' in self.__store: return int(self.__store['versions'])
        return int(self.__config.versions)

    @property
    def apps(self):
        if 'apps' in self.__store:
            if isinstance(self.__store['apps'],str):
                yield self.__store['apps']
            elif isinstance(self.__store['apps'],list):
                for app in self.__store['apps']:
                    yield app

    @property
    def __json__(self):
        ''' make RepoConfig json serializable '''
        return self.__store

    def __repr__(self): return "<RepoConfig: %s>: %s"%(self.__url,self.__store)

    #######################
    # implement hash
    #######################
    def __hash__(self):
        return self.__url.__hash__()

    #######################
    # implement "dict"
    #######################
    def __getitem__(self, key): return self.__store[key]
    def __setitem__(self, key, value): self.__store[key] = value
    def __delitem__(self, key): del self.__store[key]
    def __iter__(self): return iter(self.__store)
    def __len__(self): return len(self.__store)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from .update import Update
from .index import IndexUpdate
from .metadata import MetadataUpdate
from .apk import ApkUpdate
from .src import SrcUpdate

__all__ = ['Update','IndexUpdate','MetadataUpdate','ApkUpdate','SrcUpdate']



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from .futuressession import FuturesSessionFlex
from .verifieddownload import FuturesSessionVerifiedDownload

# disable insecure warning
# https://stackoverflow.com/questions/27981545/suppress-insecurerequestwarning-unverified-https-request-is-being-made-in-pytho
import urllib3
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

__all__ = ['FuturesSessionFlex','FuturesSessionVerifiedDownload']



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
from .futuressession import FuturesSessionFlex
from tempfile import NamedTemporaryFile
from concurrent.futures import as_completed
import os.path
import time
import shutil
import hashlib
from datetime import timedelta


logger = logging.getLogger('download.FuturesSessionVerifiedDownload')
class FuturesSessionVerifiedDownload(FuturesSessionFlex):
    BLOCKSIZE = 65536

    def __init__(self, *args, **kwargs):
        super(FuturesSessionVerifiedDownload, self).__init__(*args, **kwargs)
        self.__futures=[]

    def download(self,url,filename,timeout=600,hashType=None,hash=None):
        request = self.get(url,stream=True,timeout=timeout)
        request.filename=filename
        request.hashType=hashType
        request.hash=hash
        request.request_url=url
        self.__futures.append(request)

    @staticmethod
    def verify(filename,hashType,hash):
        h = hashlib.new(hashType)
        with open(filename, 'rb') as file:
            while True:
                byte = file.read(FuturesSessionVerifiedDownload.BLOCKSIZE)
                if not byte:
                    break
                h.update(byte)
        file_hash = h.hexdigest()
        return file_hash == hash

    def completed(self):
        for future in as_completed(self.__futures):
            url = future.request_url
            filename = future.filename
            foldername = os.path.dirname(filename)
            hashType = future.hashType
            hash = future.hash
            try:
                response = future.result()
                response.raise_for_status()
                start = time.time()
                with NamedTemporaryFile(mode='wb') as tmp:
                    for chunk in response.iter_content(chunk_size=FuturesSessionVerifiedDownload.BLOCKSIZE):
                        if chunk:
                            tmp.write(chunk)
                    tmp.flush()
                    bytes = os.stat(tmp.name).st_size
                    hbytes = FuturesSessionVerifiedDownload.h_size(bytes)
                    if hashType != None:
                        elapsed = time.time() - start
                        logger.info("downloaded %s [%s] (%s) ✔"%(response.request.url,timedelta(seconds=elapsed),hbytes))
                        tmp.seek(0)
                        if FuturesSessionVerifiedDownload.verify(tmp.name,hashType,hash):
                            if not os.path.exists(foldername):
                                os.makedirs(foldername)
                            shutil.copy(tmp.name,filename)
                            elapsed = time.time() - start
                            logger.info("hash verified %s [%s] (%s) ✔"%(response.request.url,timedelta(seconds=elapsed),hbytes))
                            yield (True,filename,bytes,hbytes,timedelta(seconds=elapsed))
                        else:
                            elapsed = time.time() - start
                            logger.warn("hash verification failed %s [%s] (%s) ❌"%(response.request.url,timedelta(seconds=elapsed),hbytes))
                            yield (False,filename,bytes,hbytes,timedelta(seconds=elapsed))
                    else:
                        if not os.path.exists(foldername):
                            os.makedirs(foldername)
                        shutil.copy(tmp.name,filename)
                        elapsed = time.time() - start
                        logger.info("downloaded %s [%s] (%s) ✔"%(response.request.url,timedelta(seconds=elapsed),hbytes))
                        yield (True,filename,bytes,hbytes,timedelta(seconds=elapsed))
            except Exception as e:
                if logging.getLogger().isEnabledFor(logging.DEBUG): logger.exception("Error downloading %s to file %s",url,filename)
                else: logger.warn("Error downloading %s to file %s: %s",url,filename,e.message if hasattr(e,'message') else str(e))
                elapsed = time.time() - start
                yield (False,filename,bytes,hbytes,timedelta(seconds=elapsed))


    #######################
    # implement "with"
    #######################
    def __enter__(self):
        super(FuturesSessionVerifiedDownload, self).__enter__()
        return self

    def __exit__(self, type, value, traceback):
        super(FuturesSessionVerifiedDownload, self).__exit__(type, value, traceback)
        self.__futures = []
        self.__files = {}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
import time
import json
import os
from datetime import timedelta
from concurrent.futures import as_completed, ThreadPoolExecutor
from ..model import Index


logger = logging.getLogger('processor.IndexFileProcessor')
class IndexFileProcessor(ThreadPoolExecutor):

    def __init__(self,*args,**kwargs):
        super(IndexFileProcessor, self).__init__(*args, **kwargs)
        self.__futures = []

    @staticmethod
    def __persist_format(file,repo):
        file.seek(0) # reset fp to beginning
        startchar = file.read(1).decode("utf-8")
        if startchar == '<': repo['format']='xml'
        else: repo['format']='json'
        file.seek(0)

    @staticmethod
    def __process(file,repo,*args):
        start = time.time()
        index=None
        IndexFileProcessor.__persist_format(file,repo)
        with file as f:
            if repo['format'] == 'json':
                index = Index.fromJSON(f,key=repo.url).monkeypatch().save(filename=repo.filename)
            else:
                index = Index.fromXML(f,key=repo.url).monkeypatch().save(filename=repo.filename)
        elapsed = time.time() - start
        return (index,timedelta(seconds=elapsed))+args

    def process(self, file, repo, *args):
        if not hasattr(file,'read'): raise IOError('file error!')
        if repo is None: raise AttributeError('repo is missing')
        args = (file,repo) + args
        self.__futures.append(self.submit(IndexFileProcessor.__process,*args))

    def completed(self):
        for future in as_completed(self.__futures):
            yield future

    #######################
    # implement "with"
    #######################
    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.shutdown()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging


logger = logging.getLogger('update.SrcUpdate')
class SrcUpdate:
    def __init__(self,config,download_timeout=600,max_workers=10):
        self.__config = config
        self.__download_timeout = download_timeout
        self.__max_workers = max_workers


    def update(self):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import collections
import os
import re
import json
import xml.etree.ElementTree as ET
from time import mktime, strptime
import logging
from urllib.parse import urljoin
from ..json import GenericJSONEncoder
from tempfile import NamedTemporaryFile
import shutil

logger = logging.getLogger('model.Index')


class Index(collections.MutableMapping):
    """
    Main Class responsible for handling the f-droid's index.xml and
    index-v1.json files.
    """

    def __init__(self, key=None, filename=None, format='json', default_locale='en-US', store={}):
        self.__key = key
        self.__filename = filename
        self.__format = format
        self.__default_locale = default_locale
        self.__store = store

    @classmethod
    def fromJSON(cls, source, **kwargs):
        if not hasattr(source, "read"):
            source = open(source, "r")
        kwargs['format'] = 'json'
        kwargs['filename'] = source.name
        return cls(**kwargs).load(source)

    @classmethod
    def fromXML(cls, source, **kwargs):
        if not hasattr(source, "read"):
            source = open(source, "r")
        kwargs['format'] = 'xml'
        kwargs['filename'] = source.name
        return cls(**kwargs).load(source)

    @property
    def filename(self):
        return str(self.__filename)

    @property
    def key(self):
        return str(self.__key)

    @property
    def repo_url(self):
        return str(self.__key)

    @property
    def format(self):
        return str(self.__format)

    @property
    def default_locale(self):
        return str(self.__default_locale)

    def findAppIds(self, key):
        if key is None:
            raise KeyError("key must not be empty")
        retVal = set()
        if key.startswith('regex:'):
            c = re.compile(key[6:], re.I | re.S)
            for app in self.__store['apps']:
                m = c.match(app['packageName'])
                if not m is None:
                    retVal.add(app['packageName'])
        elif key in ['*', '.*', 'all']:
            for app in self.__store['apps']:
                retVal.add(app['packageName'])
        else:
            for app in self.__store['apps']:
                if app == key:
                    retVal.add(app['packageName'])
        return list(retVal)

    def defaultnum(self, value):
        if value.isnumeric():
            return int(value)
        return value

    def custom_func(self, xpath):
        if not xpath is None:
            p = re.compile('(.*?):(.*)$', re.I | re.S)
            m = p.match(xpath)
            if not m is None:
                return (m.group(1), m.group(2))
            return (xpath, None)
        return (None, None)

    def run_custom_func(self, method, value):
        if method is None or value is None:
            return None
        try:
            return eval(method, {'mktime': mktime, 'strptime': strptime}, {'value': value})
        except:
            logger.exception("error in eval(%s)", method)

    def xmlattr(self, node, xml_node, key=None, xpath=None):
        if xpath is None and key is None:
            return None
        if xpath is None and not key is None:
            xpath = key
        value = self.xmlattrval(xml_node, xpath)
        if not value is None:
            if not key is None:
                node[key] = value
            else:
                node = value

    def xmlattrval(self, xml_node, xpath):
        if not xml_node is None:
            p = re.compile('^.*?\[@([^\]]+)\]$')
            m = p.match(xpath)
            if not m is None:
                attr_name = m.group(1)
            else:
                attr_name = xpath
            (xpath, func) = self.custom_func(xpath)
            if not xpath is None:
                xml_node = xml_node.find(xpath)
            if not xml_node is None:
                value = xml_node.get(attr_name, None)
                #print(xml_node,attr_name,value)
                if not value is None:
                    if not func is None:
                        return self.run_custom_func(func, self.defaultnum(value.strip()))
                    else:
                        return self.defaultnum(value.strip())
        return None

    def xmltext(self, node, xml_node, key=None, xpath=None):
        if xpath is None and key is None:
            return None
        if xpath is None and not key is None:
            xpath = key
        value = self.xmltextval(xml_node, xpath)
        if not value is None:
            if not key is None:
                node[key] = value
            else:
                node = value

    def xmltextval(self, xml_node, xpath=None):
        (xpath, func) = self.custom_func(xpath)
        if not xpath is None:
            xml_node = xml_node.find(xpath)
        if not xml_node is None:
            value = xml_node.text
            if not value is None:
                if not func is None:
                    return self.run_custom_func(func, self.defaultnum(value.strip()))
                else:
                    return self.defaultnum(value.strip())

    def xmllist(self, root, xpath):
        retVal = []
        p = re.compile('^.*?\[@([^\]]+)\]$')
        m = p.match(xpath)
        if not m is None:
            attr_name = m.group(1)
            matches = root.findall(xpath)
            if not matches is None:
                for match in matches:
                    value = match.get(attr_name, None)
                    if not value is None:
                        retVal.append(self.defaultnum(value.strip()))
        else:
            matches = root.findall(xpath)
            if not matches is None:
                for match in matches:
                    value = self.xmltextval(match)
                    if not value is None:
                        retVal.append(value)
        return retVal

    def convert(self, root):
        self.__store = {}
        store = self.__store
        repo = root.find('repo')
        if not repo is None:
            sr = store['repo'] = {}
            self.xmlattr(sr, root, 'timestamp', './repo[@timestamp]')
            self.xmlattr(sr, root, 'version', './repo[@version]')
            self.xmlattr(sr, root, 'maxage', './repo[@maxage]')
            self.xmlattr(sr, root, 'name', './repo[@name]')
            self.xmlattr(sr, root, 'icon', './repo[@icon]')
            self.xmlattr(sr, root, 'url', './repo[@address]')
            self.xmltext(sr, root, 'description', './description')
            sr['mirrors'] = self.xmllist(repo, './mirror')
        rq = store['requests'] = {}
        rq['install'] = self.xmllist(root, './install[@packageName]')
        rq['uninstall'] = self.xmllist(root, './uninstall[@packageName]')
        apps = store['apps'] = []
        pkgs = store['packages'] = {}
        applications = root.findall('application[@id]')
        for xmlapp in applications:
            appVal = {}
            self.xmltext(appVal, xmlapp, 'authorEmail', './email')
            self.xmltext(appVal, xmlapp, 'authorName', './author')
            self.xmltext(appVal, xmlapp, 'authorWebSite', './web')
            self.xmltext(appVal, xmlapp, 'bitcoin', './bitcoin')
            self.xmltext(appVal, xmlapp, 'donate', './donate')
            self.xmltext(appVal, xmlapp, 'flattr', './flattr')
            self.xmltext(appVal, xmlapp, 'liberapay', './liberapay')
            self.xmltext(appVal, xmlapp, 'litecoin', './litecoin')
            cat = self.xmltextval(xmlapp, 'categories')
            if not cat is None and len(cat) > 0:
                appVal['categories'] = cat.split(',')
            afaet = self.xmltextval(xmlapp, 'antiFeatures')
            if not afaet is None and len(afaet) > 0:
                appVal['antiFeatures'] = afaet.split(',')
            self.xmltext(appVal, xmlapp, 'suggestedVersionName',
                         './marketversion')
            self.xmltext(appVal, xmlapp, 'suggestedVersionCode',
                         './marketvercode')
            self.xmltext(appVal, xmlapp, 'issueTracker', './tracker')
            self.xmltext(appVal, xmlapp, 'changelog', './changelog')
            self.xmltext(appVal, xmlapp, 'license', './license')
            self.xmltext(appVal, xmlapp, 'name', './name')
            self.xmltext(appVal, xmlapp, 'sourceCode', './source')
            self.xmltext(appVal, xmlapp, 'webSite', './web')
            self.xmltext(appVal, xmlapp, 'added',
                         './added:int(mktime(strptime(value, "%Y-%m-%d")))')
            self.xmltext(appVal, xmlapp, 'icon', './icon')
            self.xmltext(appVal, xmlapp, 'packageName', './id')
            self.xmltext(appVal, xmlapp, 'lastUpdated',
                         './lastupdated:int(mktime(strptime(value, "%Y-%m-%d")))')
            loc = appVal['localized'] = {}
            dloc = loc[self.default_locale] = {}
            self.xmltext(dloc, xmlapp, 'description', './desc')
            self.xmltext(dloc, xmlapp, 'summary', './summary')

            appId = self.xmltextval(xmlapp, './id')
            if not appId is None:
                pkg = pkgs[appId] = []
                packages = xmlapp.findall('./package')
                for xmlpkg in packages:
                    pkgVal = {}
                    self.xmltext(
                        pkgVal, xmlpkg, 'added', './added:int(mktime(strptime(value, "%Y-%m-%d")))')
                    self.xmltext(pkgVal, xmlpkg, 'apkName', './apkname')
                    self.xmltext(pkgVal, xmlpkg, 'hash', './hash')
                    self.xmlattr(pkgVal, xmlpkg, 'hashType', './hash[@type]')
                    self.xmltext(pkgVal, xmlpkg, 'minSdkVersion', './sdkver')
                    self.xmltext(pkgVal, xmlpkg, 'targetSdkVersion',
                                 './targetSdkVersion')
                    pkgVal['packageName'] = appId
                    self.xmltext(pkgVal, xmlpkg, 'sig', './sig')
                    self.xmltext(pkgVal, xmlpkg, 'versionName', './version')
                    self.xmltext(pkgVal, xmlpkg, 'versionCode',
                                 './versioncode')
                    self.xmltext(pkgVal, xmlpkg, 'size', './size')
                    perm = xmlpkg.find('permissions')
                    if not perm is None:
                        up = pkgVal['uses-permission'] = []
                        npstr = perm.text.split(",")
                        for ps in npstr:
                            up.append(['android.permission.'+ps.strip(), None])
                    if len(pkgVal) > 0:
                        pkg.append(pkgVal)
            if len(appVal) > 1:
                apps.append(appVal)

    def monkeypatch(self):
        ''' fixup metadata paths -> url '''
        if not '_monkeypatched' in self.__store:
            if self.__key is None:
                raise AttributeError('key not defined')
            self.__store['_monkeypatched'] = True
            if 'packages' in self.__store:
                for key in self.__store['packages']:
                    for pkg in self.__store['packages'][key]:
                        pkg['apkName'] = urljoin(self.__key, pkg['apkName'])
                        if not pkg.get('srcname') is None:
                            pkg['srcname'] = urljoin(
                                self.__key, pkg['srcname'])

            if 'apps' in self.__store:
                for app in self.__store['apps']:
                    if 'icon' in app:
                        app['icon'] = urljoin(self.__key, "icons/"+app['icon'])
                    locs = app.get('localized', {})
                    for k in locs.keys():
                        loc = locs[k]
                        if 'icon' in loc:
                            loc['icon'] = urljoin(
                                self.__key, app['packageName']+'/'+k+'/'+loc['icon'])
                        if 'featureGraphic' in loc:
                            loc['featureGraphic'] = urljoin(
                                self.__key, app['packageName']+'/'+k+'/'+loc['featureGraphic'])
                        if 'promoGraphic' in loc:
                            loc['promoGraphic'] = urljoin(
                                self.__key, app['packageName']+'/'+k+'/'+loc['promoGraphic'])
                        if 'tvBanner' in loc:
                            loc['tvBanner'] = urljoin(
                                self.__key, app['packageName']+'/'+k+'/'+loc['tvBanner'])
                        if 'phoneScreenshots' in loc:
                            loc['phoneScreenshots'] = [urljoin(
                                self.__key, app['packageName']+'/'+k+'/phoneScreenshots/'+value) for value in loc['phoneScreenshots']]
                        if 'sevenInchScreenshots' in loc:
                            loc['sevenInchScreenshots'] = [urljoin(
                                self.__key, app['packageName']+'/'+k+'/sevenInchScreenshots/'+value) for value in loc['sevenInchScreenshots']]
                        if 'tenInchScreenshots' in loc:
                            loc['tenInchScreenshots'] = [urljoin(
                                self.__key, app['packageName']+'/'+k+'/tenInchScreenshots/'+value) for value in loc['tenInchScreenshots']]
                        if 'tvScreenshots' in loc:
                            loc['tvScreenshots'] = [urljoin(
                                self.__key, app['packageName']+'/'+k+'/tvScreenshots/'+value) for value in loc['tvScreenshots']]
                        if 'wearScreenshots' in loc:
                            loc['wearScreenshots'] = [urljoin(
                                self.__key, app['packageName']+'/'+k+'/wearScreenshots/'+value) for value in loc['wearScreenshots']]
        return self

    def load(self, file, format=None):
        if not hasattr(file, 'read'):
            file = open(file, 'r')
        if not format is None:
            self.__format = format

        if self.__format == 'json':
            with file as idxfl:
                self.__store = json.load(idxfl)
        elif self.__format == 'xml':
            # we have no json verison we need som transformation to apply
            tree = ET.parse(file)
            file.close()
            root = tree.getroot()
            self.convert(root)
            self.__format = 'json'
        else:
            # try loading json anyway
            try:
                self.load(file, format='json')
            except:
                self.load(file, format='xml')
        return self

    def save(self, filename=None):
        if not filename is None:
            self.__filename = filename
        with NamedTemporaryFile(mode='w') as tmp:
            json.dump(self.__store, tmp, sort_keys=True,
                      indent=4, cls=GenericJSONEncoder)
            tmp.flush()
            shutil.copy(tmp.name, self.__filename)
        return self

    def __repr__(self):
        return "<Index: %s>" % str(json.dumps(self.__store, indent=4))

    @property
    def __json__(self):
        ''' make Index json serializable '''
        return self.__store
    #######################
    # implement "dict"
    #######################

    def __getitem__(self, key):
        return self.__store[key]

    def __setitem__(self, key, value):
        self.__store[key] = value

    def __delitem__(self, key):
        del self.__store[key]

    def __iter__(self):
        return iter(self.__store)

    def __len__(self):
        return len(self.__store)
    #######################
    # implement "with"
    #######################

    def __enter__(self):
        self.load(self.__filename)
        return self

    def __exit__(self, type, value, traceback):
        self.save()



"""main entrypoint into fdroid-dl."""
import logging
import click
from .model import Config
from .update import Update


name = 'fdroid-dl'
version = '0.0.1'
author = 't4skforce'
author_mail = '7422037+t4skforce@users.noreply.github.com'

logging.basicConfig(level=logging.INFO,
                    format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
logger = logging.getLogger(name)


@click.group(invoke_without_command=True)
@click.option('-d', '--debug', is_flag=True, default=False, help='enable debug level logging')
@click.option('-c', '--config', default='fdroid-dl.json', type=click.Path(dir_okay=False, writable=True, resolve_path=True), show_default=True, help='location of your fdroid-dl.json configuration file')
@click.option('-r', '--repo', default='./repo', type=click.Path(file_okay=False, writable=True, resolve_path=True), show_default=True, help='location of your fdroid repository to store the apk files')
@click.option('-m', '--metadata', default='./metadata', type=click.Path(file_okay=False, writable=True, resolve_path=True), show_default=True, help='location of your fdroid metadata to store the asset files')
@click.option('--cache', default='./.cache', type=click.Path(file_okay=False, writable=True, resolve_path=True), show_default=True, help='location for fdroid-dl to store cached data')
@click.pass_context
def cli(ctx, debug, config, repo, metadata, cache):
    """
        Is a python based f-droid mirror generation and update utility.
        Point at one or more existing f-droid repositories and the utility will download the metadata (pictures, descriptions,..)
        for you and place it in your local system.

        Simply run "fdroid-dl update && fdroid update" in your folder with repo and you are set.
    """
    if ctx.obj is None:
        ctx.obj = {}
    ctx.obj['debug'] = debug
    ctx.obj['config'] = config
    ctx.obj['repo'] = repo
    ctx.obj['metadata'] = metadata
    ctx.obj['cache_dir'] = cache
    if debug:
        logger.setLevel(logging.DEBUG)
    logger.info('Debug mode is %s' % ('on' if debug else 'off'))
    if ctx.invoked_subcommand is None:
        with click.Context(cli) as ctx:
            click.echo(cli.get_help(ctx))


@cli.group(name='update', invoke_without_command=True, short_help='starts updating process')
@click.option('--index/--no-index', default=True, show_default=True, help='download repository index files')
@click.option('--metadata/--no-metadata', default=True, show_default=True, help='download metadata assset files')
@click.option('--apk/--no-apk', default=True, show_default=True, help='download apk files')
@click.option('--src/--no-src', default=True, show_default=True, help='download src files')
@click.pass_context
def update(ctx, index, metadata, apk, src):
    with Config(ctx.obj['config'], cache_dir=ctx.obj['cache_dir']) as cfg:
        u = Update(cfg)
        if index:
            u.index()
        if metadata:
            u.metadata()
        if apk:
            u.apk()
        if src:
            u.src()


__all__ = ['cli', 'update']

if __name__ == '__main__':
    cli()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
import requests


logger = logging.getLogger('update.Selector')
class Selector:
    def __init__(self,config):
        self.__config = config

    def __metaRepos(self):
        for repo in self.__config.repos:
            if not 'error' in repo: # ignore repos with download error
                yield repo

    def allApps(self,dupes=False,session=None):
        yielded = set()
        for repo in self.__metaRepos():
            self.applySessionSettings(repo, session)
            for selector in repo.apps:
                for appid in repo.index.findAppIds(selector):
                    if not yielded in yielded or dupes:
                        yielded.add(appid)
                        yield (repo,appid)

    def applySessionSettings(self, repo, session):
        ''' apply security settings for basic auth and ssl verification '''
        if not session is None:
            if not repo.auth is None or repo.verify == False:
                ts = requests.Session()
                ts.auth=repo.auth
                ts.verify=repo.verify
                session.map(repo.url,ts)



import unittest
import fdroid_dl
from unittest.mock import patch


class ConfigTestSuite(unittest.TestCase):

    @patch('os.makedirs')
    @patch.object(fdroid_dl.Config, 'save')
    def test_prepare_dirs(self, save, makedirs):
        print(save, makedirs)
        with fdroid_dl.Config() as c:
            makedirs.assert_any_call('./repo', exist_ok=True)
            makedirs.assert_any_call('./metadata', exist_ok=True)
            makedirs.assert_any_call('.cache', exist_ok=True)
            self.assertEqual(c.filename, 'fdroid-dl.json')
            self.assertEqual(c.repo_dir, './repo')
            self.assertEqual(c.metadata_dir, './metadata')
            self.assertEqual(c.cache_dir, '.cache')
            self.assertIsNot(c.get('f-droid'), None)
            self.assertEqual([x for x in iter(c)],
                             [x for x in iter(fdroid_dl.Config.DEFAULTS)])
            assert not save.called
        assert save.called



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
import collections
import copy
import json
from ..json import GenericJSONEncoder


logger = logging.getLogger('model.AppMetadata')
class AppMetadata(collections.MutableMapping):
    def __init__(self, appid, cfg={}, json={}, default_locale='en-US'):
        if appid is None: raise KeyError("no appid defined")
        self.__id = appid
        self.__cfg = cfg
        if not json is None: self.__store = self.__merge(copy.deepcopy(cfg),copy.deepcopy(json))
        else: self.__store = copy.deepcopy(cfg)
        self.__default_locale = default_locale

    def __merge(self,source,destination):
        for key, value in source.items():
            if isinstance(value, dict):
                node = destination.setdefault(key, {})
                self.__merge(value, node)
            else:
                destination[key] = value
        return destination

    @property
    def id(self):
        return str(self.__id)

    @property
    def appid(self):
        return str(self.__id)

    def merge(self,appmetadata):
        ''' used for adding addition data from e.g. index '''
        if not isinstance(appmetadata,AppMetadata) and not isinstance(appmetadata,dict):
            raise NotImplementedError("I can only merge two instances of AppMetadata at the moment")
        self.__store = self.__merge(self.__store,copy.deepcopy(appmetadata))
        return self

    def update(self,appmetadata):
        ''' apply manual changes and save tose in config '''
        if not isinstance(appmetadata,AppMetadata) and not isinstance(appmetadata,dict):
            raise NotImplementedError("I can only merge two instances of AppMetadata at the moment")
        self.__store = self.__merge(copy.deepcopy(appmetadata),self.__store)
        self.__cfg = self.__merge(copy.deepcopy(appmetadata),self.__cfg)
        return self

    def localized(self,locale=None):
        return self.__store.get('localized',{}).get(locale,{})

    def full_description(self,locale=None):
        return self.localized(locale).get('description',None)

    def short_description(self,locale=None):
        return self.localized(locale).get('summary',None)

    def title(self,locale=None):
        return self.localized(locale).get('name',None)

    def featureGraphic(self,locale=None):
        return self.localized(locale).get('featureGraphic',None)

    def icon(self,locale=None):
        return self.localized(locale).get('icon',None)

    def promoGraphic(self,locale=None):
        return self.localized(locale).get('promoGraphic',None)

    def tvBanner(self,locale=None):
        return self.localized(locale).get('tvBanner',None)

    def phoneScreenshots(self,locale=None):
        return self.localized(locale).get('phoneScreenshots',[])

    def sevenInchScreenshots(self,locale=None):
        return self.localized(locale).get('sevenInchScreenshots',[])

    def tenInchScreenshots(self,locale=None):
        return self.localized(locale).get('tenInchScreenshots',[])

    def tvScreenshots(self,locale=None):
        return self.localized(locale).get('tvScreenshots',[])

    def wearScreenshots(self,locale=None):
        return self.localized(locale).get('wearScreenshots',[])

    @property
    def locales(self):
        if 'localized' in self:
            return list(self['localized'].keys())
        return []

    def __repr__(self):
        return "<AppMetadata: %s>"%str(json.dumps(self.__store, indent=4, cls=GenericJSONEncoder))

    @property
    def __json__(self):
        ''' make Metadata json serializable '''
        return self.__cfg

    #######################
    # implement hash
    #######################
    def __hash__(self):
        return self.__id.__hash__()

    #######################
    # implement "dict"
    #######################
    def __getitem__(self, key):
        return self.__store[key]
    def __setitem__(self, key, value):
        self.__cfg[key] = value
        self.__store[key] = value
    def __delitem__(self, key):
        if key in self.__cfg: del self.__cfg[key]
        del self.__store[key]
    def __iter__(self):
        return iter(self.__store)
    def __len__(self):
        return len(self.__store)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
    def repo(self):
        with Downloader(threads=self.config.threads) as d:
            downloads = {}
            for repo in self.config.repos:
                try:
                    if not 'error' in repo: # ignore repos with download error
                        for appselector in repo.apps:
                            for appid in repo.index.findAppIds(appselector):
                                packages = repo.index.get('packages',{}).get(appid,[])
                                if not appid in downloads: downloads[appid]=[]
                                for pkg in packages:
                                    if not pkg.get('apkName') is None and not pkg.get('hash') is None and not pkg.get('hashType') is None:
                                        downloads[appid].append(pkg)
                except Exception as e:
                    logger.error("error processing repo metadata",e)
            for appid in downloads.keys():
                packages = sorted(downloads[appid], key=lambda e:e.get('versionCode',0), reverse=True)
                for idx,pkg in enumerate(packages):
                    if idx >= self.config.versions: break;
                    filename = os.path.basename(str(urlparse(pkg.get('apkName')).path))
                    #filename = "%s_%s.apk"%(appid,pkg.get('versionCode',pkg.get('versionName')))
                    filepath=os.path.join(self.config.repo_dir,filename)
                    if not os.path.exists(filepath):
                        d.add(pkg.get('apkName'), filename=filepath[:-4]+'.tmp', ref=pkg ,cached=repo.hash)

            for result,pkg in d.results():
                (dl_url,filename,hash,bytes,hbytes) = result
                logger.info("downloaded %s (%s) ✔"%(filename,hbytes))
                file_hash = self.__hash_file(filename,pkg.get('hashType'))
                if file_hash == pkg.get('hash'):
                    logger.info("hash verified %s (%s) ✔"%(filename,hbytes))
                    os.rename(filename,filename[:-4]+'.apk')
                else:
                    if os.path.isfile(filename): os.remove(filename)
                    logger.warn("removed %s (%s) hash verification failed ❌"%(filename,hbytes))

            if self.config.src_download == True:
                for appid in downloads.keys():
                    packages = sorted(downloads[appid], key=lambda e:e.get('versionCode',0), reverse=True)
                    for idx,pkg in enumerate(packages):
                        if idx >= self.config.versions: break;
                        if not pkg.get('srcname') is None:
                            filename = os.path.basename(urlparse(pkg.get('srcname')).path)
                            filepath=os.path.join(self.config.repo_dir,filename)
                            if not os.path.exists(filepath):
                                d.add(pkg.get('srcname'), filename=filepath, ref=pkg ,cached=repo.hash)
                for result,pkg in d.results():
                    (dl_url,filename,hash,bytes,hbytes) = result
                    logger.info("downloaded %s (%s) ✔"%(filename,hbytes))


        return self
'''



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
import time
import os.path
import time
from datetime import timedelta
from urllib.parse import urlparse
from .selector import Selector
from ..download import FuturesSessionVerifiedDownload, FuturesSessionFlex

logger = logging.getLogger('update.ApkUpdate')
class ApkUpdate(Selector):
    def __init__(self,config,download_timeout=600,max_workers=10):
        super(ApkUpdate,self).__init__(config)
        self.__config = config
        self.__download_timeout = download_timeout
        self.__max_workers = max_workers

    def allPackage(self,session=None):
        downloads = {}
        logging.info("collecting apps to download")
        start = time.time()
        apkcnt = 0
        # search pkgs
        for repo,appid in self.allApps(dupes=True):
            self.applySessionSettings(repo, session)
            packages = repo.index.get('packages',{}).get(appid,[])
            if not appid in downloads: downloads[appid]=[]
            for pkg in packages:
                if not pkg.get('apkName') is None and not pkg.get('hash') is None and not pkg.get('hashType') is None:
                    downloads[appid].append(pkg)
                    apkcnt += 1
        elapsed = time.time() - start
        logging.info("found (%s) apk files to download (%s)"%(apkcnt,timedelta(seconds=elapsed)))

        for appid in downloads.keys():
            # sort newest first
            packages = sorted(downloads[appid], key=lambda e:e.get('versionCode',0), reverse=True)
            for idx,pkg in enumerate(packages):
                if idx >= self.__config.versions: break # only download number of confgured files
                url = pkg.get('apkName')
                filename = os.path.basename(str(urlparse(url).path))
                filepath = os.path.join(self.__config.repo_dir,filename)
                yield (url, filepath, pkg.get('hash'), pkg.get('hashType'))

    def update(self):
        meta = self.__config.metadata
        logger.info("UPDATING apk files")
        start = time.time()
        cnt = 0
        ecnt = 0
        with FuturesSessionVerifiedDownload(max_workers=self.__max_workers) as session:
            for url, filename, hash, hashType in self.allPackage(session=session):
                if os.path.exists(filename):
                    start2 = time.time()
                    if FuturesSessionVerifiedDownload.verify(filename,hashType,hash):
                        elapsed = time.time() - start2
                        logger.info("hash verified %s [%s] (%s) ✔"%(os.path.basename(filename),timedelta(seconds=elapsed),FuturesSessionFlex.h_size(os.stat(filename).st_size)))
                    else:
                        session.download(url,filename,timeout=self.__download_timeout,hash=hash,hashType=hashType)
                else:
                    session.download(url,filename,timeout=self.__download_timeout,hash=hash,hashType=hashType)

            dlsum = 0
            for ok,filename,bytes,hbytes,elapsed in session.completed():
                if ok:
                    cnt += 1
                    dlsum += bytes
                else:
                    ecnt += 1
        elapsed = time.time() - start
        logger.info("UPDATED apk files, files(%s) errors(%s) [%s] (%s)"%(cnt,ecnt,FuturesSessionFlex.h_size(dlsum),timedelta(seconds=elapsed)))



"""Package exposing fdroid_dl.Config."""
# -*- coding: utf-8 -*-

import logging
import collections
import copy
import json
import os
import os.path
import shutil
from tempfile import NamedTemporaryFile
from .repoconfig import RepoConfig
from .metadata import Metadata
from .index import Index
from ..json import GenericJSONEncoder


LOGGER = logging.getLogger('model.Config')


class Config(collections.MutableMapping):
    """
    Main Class responsible for handling the f-droid.json config file.

    This Class represents the config file used by fdroid-dl and will default to
    https://f-droid.org/repo/ as bas repo and add at least the
    **org.fdroid.fdroid** app for download into the queue.
    """

    DEFAULTS = {
        "f-droid": {
            'https://f-droid.org/repo/': {
                "apps": ["org.fdroid.fdroid"]
            }
        },
        "metadata": {}
    }

    def __init__(self, filename='fdroid-dl.json', repo_dir='./repo',
                 metadata_dir='./metadata', cache_dir='.cache'):
        """
        Parameters
        ----------
        filename : str
            filename including path where to store the generated config file
        repo_dir : str
            path to f-droid repo, apk files will be downloded here
        metadata_dir : str
            path to f-droid metadata directory, images and descriptions will
            be stored here
        cache_dir : str
            directory to store extracted and parsed index.json files from
            f-droid repo
        """
        self.__filename = filename
        self.__repo = repo_dir
        self.__metadata_dir = metadata_dir
        self.__cache_dir = cache_dir
        self.__store = {}
        self.__indices = {}
        self.__metadata = None
        self.__init_defaults()
        self.__prepare_fs()

    @property
    def filename(self):
        return str(self.__filename)

    @property
    def repo_dir(self):
        return str(self.__repo)

    @property
    def metadata_dir(self):
        return str(self.__metadata_dir)

    @property
    def cache_dir(self):
        return str(self.__cache_dir)

    @property
    def store(self):
        return str(self.__store)

    @property
    def size(self):
        return len(self.__store.keys())

    def __init_defaults(self):
        self.__store = copy.deepcopy(Config.DEFAULTS)
        for key in Config.DEFAULTS['f-droid'].keys():
            cfg = RepoConfig(key, self.__store['f-droid'][key], self)
            self.__store['f-droid'][cfg.url] = cfg
        self.__metadata = Metadata(self, Config.DEFAULTS.get('metadata', {}))

    def __prepare_fs(self):
        os.makedirs(self.__repo, exist_ok=True)
        os.makedirs(self.__metadata_dir, exist_ok=True)
        os.makedirs(self.__cache_dir, exist_ok=True)

    def load(self, file=None):
        """
        Loads given file handler or filename into Config object and resolves.

        """
        if file is None:
            file = self.__filename
        if not hasattr(file, 'read'):
            file = open(file, 'r')

        if os.path.isfile(self.__filename):
            with open(self.__filename) as config_file:
                try:
                    file_data = json.load(config_file)
                    if 'f-droid' in file_data:
                        for key in file_data['f-droid'].keys():
                            cfg = RepoConfig(
                                key, file_data['f-droid'][key], self)
                            self.__store['f-droid'][cfg.url] = cfg
                    if 'metadata' in file_data:
                        self.__metadata = Metadata(self, file_data['metadata'])
                        self.__store['metadata'] = self.__metadata
                except Exception:
                    LOGGER.exception("Fatal error reading %s", config_file)

    def save(self):
        print("save")
        with NamedTemporaryFile(mode='w') as tmp:
            json.dump(self.__store, tmp, sort_keys=True,
                      indent=4, cls=GenericJSONEncoder)
            tmp.flush()
            shutil.copy(tmp.name, self.__filename)
        return self

    @property
    def repos(self):
        for key in self.__store['f-droid'].keys():
            cfg = self.__store['f-droid'][key]
            if not isinstance(cfg, RepoConfig):
                self.__store['f-droid'][key] = RepoConfig(key, cfg, self)
            yield cfg

    @property
    def indices(self):
        for repo in self.repos:
            if repo.url in self.__indices:
                yield self.__indices[repo.url]
            elif os.path.exists(repo.filename):
                with Index.fromJSON(repo.filename, key=repo.url) as idx:
                    self.__indices[repo.url] = idx
                    yield idx

    def repo(self, url):
        """
        Return RepoConfig based on url.

        Searches in config file for given url and returns the corresponding
        RepoConfig Class representation of it. if the repo is marked as
        having errors it will not be found. A KeyError is being raised if
        given url is not found or has an error node in the json file.

        Parameters
        ----------
        key : str
            url of repo in config file

        Returns
        -------
        fdroid_dl.model.Index
            index found for given key

        Raises
        ------
        KeyError
            raised if given url is not found or is in error state

        """
        if url in self.__store['f-droid']:
            cfg = self.__store['f-droid'][url]
            if not isinstance(cfg, RepoConfig):
                self.__store['f-droid'][url] = RepoConfig(url, cfg, self)
            if 'error' not in cfg:
                return cfg
        raise KeyError("repo with url: %s not found" % url)

    def index(self, url):
        """
        Return Index based on url.

        Searches for downloaded index file based on repository url as
        configured in config file.

        Parameters
        ----------
        url : str
            url of repo in config file

        Returns
        -------
        fdroid_dl.model.Index
            index found for given url

        """
        repo = self.repo(url)
        if repo.url in self.__indices:
            return self.__indices[repo.url]
        if os.path.exists(repo.filename):
            with Index.fromJSON(repo.filename, key=repo.url) as idx:
                self.__indices[repo.url] = idx
            return self.__indices[repo.url]
        raise KeyError(
            "index with url: %s not found on filesystem file: %s" %
            (url, repo.filename))

    def __repr__(self):
        """."""
        return "<Config: %s>" % str(self.__store)
    #######################
    # implement "dict"
    #######################

    def __getitem__(self, key):
        """."""
        return self.__store[key]

    def __setitem__(self, key, value):
        """."""
        self.__store[key] = value

    def __delitem__(self, key):
        """."""
        del self.__store[key]

    def __iter__(self):
        """."""
        return iter(self.__store)

    def __len__(self):
        """."""
        return len(self.__store)
    #######################
    # implement "with"
    #######################

    def __enter__(self):
        """."""
        self.load()
        return self

    def __exit__(self, type, value, traceback):
        """."""
        self.save()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
import time
from datetime import timedelta
import os.path
import yaml
from .selector import Selector
from ..download import FuturesSessionVerifiedDownload
from urllib.parse import urlparse
import requests

logger = logging.getLogger('update.MetadataUpdate')
class MetadataUpdate(Selector):
    def __init__(self,config,download_timeout=600,max_workers=10):
        super(MetadataUpdate,self).__init__(config)
        self.__config = config
        self.__download_timeout = download_timeout
        self.__max_workers = max_workers
        self.__loaded = False
        self.__loadAll()

    def __loadAll(self):
        if not self.__is_loaded():
            self.__config.metadata.loadAll()
            self.__loaded = True

    def __is_loaded(self):
        return self.__loaded == True

    def _setyamlattr(sefl,yaml_key,json_key,yaml_data,app_meta):
        value = app_meta.get(json_key)
        if not value is None:
            yaml_data[yaml_key]=value

    def updateYAML(self):
        meta = self.__config.metadata
        logger.info("UPDATING YAML metadata")
        start = time.time()
        cnt = 0
        for repo,appid in self.allApps():
            try:
                app_meta = meta[appid]
                yaml_file = os.path.join(self.__config.metadata_dir,appid+'.yml')
                yaml_data = {}
                if os.path.exists(yaml_file):
                    with open(yaml_file, 'r') as yfl:
                        yaml_data = yaml.load(yfl)
                    if yaml_data is None:
                        yaml_data = {}
                self._setyamlattr('Categories','categories',yaml_data,app_meta)
                self._setyamlattr('AuthorName','authorName',yaml_data,app_meta)
                self._setyamlattr('AuthorEmail','authorEmail',yaml_data,app_meta)
                self._setyamlattr('License','license',yaml_data,app_meta)
                #self._setyamlattr('Name','name',yaml_data,app_meta)
                self._setyamlattr('WebSite','webSite',yaml_data,app_meta)
                self._setyamlattr('SourceCode','sourceCode',yaml_data,app_meta)
                self._setyamlattr('IssueTracker','issueTracker',yaml_data,app_meta)
                self._setyamlattr('Changelog','changelog',yaml_data,app_meta)
                self._setyamlattr('Donate','donate',yaml_data,app_meta)
                self._setyamlattr('FlattrID','flattr',yaml_data,app_meta)
                self._setyamlattr('LiberapayID','liberapay',yaml_data,app_meta)
                self._setyamlattr('Bitcoin','bitcoin',yaml_data,app_meta)
                self._setyamlattr('Litecoin','litecoin',yaml_data,app_meta)
                self._setyamlattr('AntiFeatures','antiFeatures',yaml_data,app_meta)
                with open(yaml_file,'w') as stream:
                    yaml.dump(yaml_data,stream,default_flow_style=False)
                cnt+=1
            except Exception as e:
                if logging.getLogger().isEnabledFor(logging.DEBUG): logger.exception()
                else: logger.warn(str(e.message) if hasattr(e,'message') else str(e))
        elapsed = time.time() - start
        logger.info("UPDATED YAML metadata, %s files (%s)"%(cnt,timedelta(seconds=elapsed)))

    def __write_text(self,text,filename,foldername):
        if not text is None:
            text = str(text)
            if not os.path.exists(foldername):
                os.makedirs(foldername)
            with open(os.path.join(foldername,filename), "w") as text_file:
                text_file.write(text)

    def __download_image(self,url,filename,session):
        if not url is None:
            session.download(url,filename,timeout=self.__download_timeout)

    def __download_images(self,urls,foldername,session):
        for url in urls:
            filename = os.path.basename(urlparse(url).path)
            filename = os.path.join(foldername,filename)
            self.__download_image(url,filename,session)

    def updateAssets(self):
        meta = self.__config.metadata
        logger.info("UPDATING Assets metadata")
        start = time.time()
        cnt = 0
        ecnt = 0
        with FuturesSessionVerifiedDownload(max_workers=self.__max_workers) as session:
            for repo,appid in self.allApps(session=session):
                try:
                    loc_appid = os.path.join(self.__config.metadata_dir,appid)
                    app_meta = meta[appid]
                    for locale in app_meta.locales:
                        loc_path = os.path.join(loc_appid,locale)
                        imags_path = os.path.join(loc_path,'images')
                        # TODO: chech if we need to download really all images & clean folder before/after?
                        self.__write_text(app_meta.full_description(locale),'full_description.txt',loc_path)
                        self.__write_text(app_meta.short_description(locale),'short_description.txt',loc_path)
                        self.__write_text(app_meta.title(locale),'title.txt',loc_path)

                        self.__download_image(app_meta.icon(locale),os.path.join(imags_path,'icon.png'),session)
                        self.__download_image(app_meta.featureGraphic(locale),os.path.join(imags_path,'featureGraphic.png'),session)
                        self.__download_image(app_meta.promoGraphic(locale),os.path.join(imags_path,'promoGraphic.png'),session)
                        self.__download_image(app_meta.tvBanner(locale),os.path.join(imags_path,'tvBanner.png'),session)

                        self.__download_images(app_meta.phoneScreenshots(locale),os.path.join(loc_path,'phoneScreenshots'),session)
                        self.__download_images(app_meta.sevenInchScreenshots(locale),os.path.join(loc_path,'sevenInchScreenshots'),session)
                        self.__download_images(app_meta.tenInchScreenshots(locale),os.path.join(loc_path,'tenInchScreenshots'),session)
                        self.__download_images(app_meta.tvScreenshots(locale),os.path.join(loc_path,'tvScreenshots'),session)
                        self.__download_images(app_meta.wearScreenshots(locale),os.path.join(loc_path,'wearScreenshots'),session)
                except Exception as e:
                    logger.exception("Error processing Asset download for",appid)
            for ok,filename,bytes,hbytes,elapsed in session.completed():
                if ok: cnt+=1
                else: ecnt+=1
        elapsed = time.time() - start
        logger.info("UPDATED Assets metadata, %s files, %s errors (%s)"%(cnt,ecnt,timedelta(seconds=elapsed)))



#!/usr/bin/env python
# coding: utf-8
import os.path
import warnings
import sys
import setuptools
import fdroid_dl

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name=fdroid_dl.name,
    version=fdroid_dl.version,
    author=fdroid_dl.author,
    author_email=fdroid_dl.author_mail,
    description="fdroid-dl is a f-droid (offline) mirror generation and update utility",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/t4skforce/fdroid-dl",
    entry_points={'console_scripts': ['fdroid-dl=fdroid_dl:cli']},
    packages=setuptools.find_packages(),
    install_requires=[
        'requests-futures>=0.9.7',
        'PyYAML>=3.13',
        'click>=6.7'
    ],
    python_requires='>=3.5',
    classifiers=(
        "Development Status :: 2 - Pre-Alpha",
        "Environment :: Console",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ),
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
import os.path
from concurrent.futures import as_completed
import requests
from ..download import FuturesSessionFlex
from ..processor import IndexFileProcessor


logger = logging.getLogger('update.IndexUpdate')
class IndexUpdate:
    def __init__(self,config,head_timeout=10,index_timeout=60,max_workers=10):
        self.config = config
        self.head_timeout = head_timeout
        self.index_timeout = index_timeout
        self.max_workers = max_workers

    def required(self,repos,timeout=60):
        # try new index
        head_futures = self.__future(repos=repos,timeout=timeout)
        # process result
        (new_index,notfound)=self.__head_response(head_futures)
        # retry old index
        head_futures = self.__future(repos=notfound,attr='url_index',timeout=timeout)
        # process result
        (old_index,notfound)=self.__head_response(head_futures)
        # new_index[] needs to be fetched with index-v1.jar
        # old_index[] needs to be fetched with index.jar
        return (new_index,old_index)

    def download(self,new_index,old_index,timeout=60):
        if new_index is None: raise AttributeError('new_index missing %s'%repos)
        if old_index is None: raise AttributeError('old_index missing %s'%repos)
        new_futures = self.__future(repos=new_index,timeout=timeout,http_method=FuturesSessionFlex.get,background_callback=FuturesSessionFlex.extract_jar, stream=True)
        old_futures = self.__future(repos=old_index,timeout=timeout,http_method=FuturesSessionFlex.get,background_callback=FuturesSessionFlex.extract_jar,attr='url_index', stream=True)
        self.__download_response(new_futures)
        self.__download_response(old_futures)

    def __future(self, repos=None, attr='url_index_v1', http_method=FuturesSessionFlex.head, background_callback=None, timeout=60,**kwargs):
        if repos is None: raise AttributeError('repos missing %s'%repos)
        if background_callback is None:
            background_callback=FuturesSessionFlex.add_hash
        futures=[]
        with FuturesSessionFlex(max_workers=self.max_workers) as session:
            for repo in repos:
                if not repo.auth is None or repo.verify == False:
                    ts = requests.Session()
                    ts.auth=repo.auth
                    ts.verify=repo.verify
                    session.map(getattr(repo,attr),ts)
                request = http_method(session,getattr(repo,attr), background_callback=background_callback, timeout=timeout, **kwargs)
                request.repo=repo # pass repo ref to future processing
                futures.append(request)
        return futures

    def __as_completed(self,futures):
        for future in as_completed(futures):
            repo = future.repo
            try:
                response = future.result()
                response.raise_for_status()
                yield (repo,response,True)
            except requests.exceptions.HTTPError as e:
                if e.response.status_code == 404: # try old index file
                    if logging.getLogger().isEnabledFor(logging.DEBUG): logger.exception()
                    else: logger.warn(str(e))
                    yield (repo,response,False)
                else:
                    repo['error']={'code':e.response.status_code if isinstance(e, exceptions.HTTPError) else 600,'msg':str(e)}
                    logger.exception()

    def __head_response(self,futures,timeout=60):
        success=[]
        notfound=[]
        for repo,response,ok in self.__as_completed(futures):
            if ok:
                # check cache
                logger.info("HEAD %s (%s) "%(response.url,response.elapsed))
                if not os.path.exists(repo.filename) or repo.hash != response.hash:
                    if 'error' in repo:
                        del repo['error']
                    repo['hash']=response.hash
                    success.append(repo)
                    if not os.path.exists(repo.filename):
                        logger.warn("CACHE - (miss) - %s.cache file not found!"%(repo.id,))
                    else:
                        logger.info("CACHE - (miss) - %s - %s)"%(repo.key,response.hash))
                else:
                    # skip do nothing for cache hits
                    logger.info("CACHE - (hit) - %s - %s)"%(repo.key,response.hash))
            else:
                notfound.append(repo)
        return (success,notfound)

    def __download_response(self,futures,timeout=60):
        with IndexFileProcessor(max_workers=self.max_workers) as ifp:
            for repo,response,ok in self.__as_completed(futures):
                if response.ok:
                    logger.info("DOWNLOADED %s [%s] (%s) "%(response.url,response.elapsed,response.h_size))
                    idx = response.index
                    ifp.process(response.index,repo,repo.url,response.h_size)
            for future in ifp.completed():
                (index,elapsed,url,h_size) = future.result()
                repo['name']=index.get('repo',{}).get('name')
                logger.info("UPDATED %s - %s [%s] (%s) "%(repo['name'],url,elapsed,h_size))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
from requests_futures.sessions import FuturesSession
import hashlib
import time
import copy
from datetime import timedelta
from zipfile import ZipFile
from tempfile import NamedTemporaryFile
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry
import os

logger = logging.getLogger('download.FuturesSessionFlex')
class FuturesSessionFlex(FuturesSession):
    BLOCKSIZE = 65536
    SUFFIXES = ['B', 'KB', 'MB', 'GB', 'TB', 'PB']

    def __init__(self, max_workers=1, user_agent='Mozilla/5.0 (compatible; Googlebot/2.1; +http://www.google.com/bot.html)', *args, **kwargs):
        kwargs.update({'max_workers':max_workers})
        super(FuturesSessionFlex, self).__init__(*args, **kwargs)
        self.__sessions={}
        self.__sessions_keys=[]

        self.__fs_kwargs={}
        self.__fs_kwargs.update(kwargs)

        _adapter_kwargs = {'pool_connections': max_workers,'pool_maxsize': max_workers,'pool_block':True}
        self.mount('https://', HTTPAdapter(**_adapter_kwargs))
        self.mount('http://', HTTPAdapter(**_adapter_kwargs))
        if self.headers is None:
            self.headers={}
        self.headers.update({'User-Agent': user_agent})

    def map(self,pattern='http://',session=None):
        ''' if called with session None -> default session for ctor is used '''
        kwargs = copy.deepcopy(self.__fs_kwargs)
        kwargs['session']=session
        if not pattern in self.__sessions: self.__sessions_keys.append(pattern)
        self.__sessions[pattern]=FuturesSessionFlex(*(), **kwargs)
        self.__sessions_keys=sorted(self.__sessions_keys, key=len, reverse=True)

    def set_headers(self,headers):
        self.headers.update(headers)

    @staticmethod
    def h_size(nbytes):
        i = 0
        while nbytes >= 1024 and i < len(FuturesSessionFlex.SUFFIXES)-1:
            nbytes /= 1024.
            i += 1
        f = ('%.2f' % nbytes).rstrip('0').rstrip('.')
        return '%s %s' % (f, FuturesSessionFlex.SUFFIXES[i])

    @staticmethod
    def add_size(session, response):
        if 'Content-Length' in response.headers:
            response.size=int(response.headers.get('Content-Length',0))
        else:
            logger.warn("Content-Length Header not provided by %s"%response.url)
            response.size=len(response.content)
        response.h_size=FuturesSessionFlex.h_size(response.size)

    @staticmethod
    def add_hash(session, response):
        response.hash=None
        if response.ok:
            ts = str(time.time())
            cv = response.headers.get('Last-Modified',ts)+response.headers.get('ETag',ts)
            response.hash=hashlib.sha1(str(cv).encode('UTF-8')).hexdigest()

    @staticmethod
    def extract_jar(session, response):
        if response.ok:
            start = time.time()
            FuturesSessionFlex.add_size(session, response)
            response.index=NamedTemporaryFile()
            with NamedTemporaryFile() as f:
                for chunk in response.iter_content(chunk_size=FuturesSessionFlex.BLOCKSIZE):
                    if chunk: f.write(chunk)
                zip_file = ZipFile(f)
                idxfile = 'index-v1.json' if 'index-v1.json' in zip_file.namelist() else 'index.xml'
                with zip_file.open(idxfile) as file:
                    while True:
                        byte = file.read(FuturesSessionFlex.BLOCKSIZE)
                        if not byte: break
                        response.index.write(byte)
            logging.debug("%s - %s - (%s)"%(response.index.name,response.url,FuturesSessionFlex.h_size(os.stat(response.index.name).st_size)))
            elapsed = time.time() - start
            response.elapsed+=timedelta(seconds=elapsed)

    def __lookup_fs_session(self,url):
        # fast direct matches
        if url in self.__sessions:
            return self.__sessions[url]
        # slower pattern search depends on pattern count and size
        for k in self.__sessions_keys:
            if url.find(k) == 0:
                return self.__sessions[k]
        return None

    def request(self, *args, **kwargs):
        session = self.__lookup_fs_session(args[1])
        if not session is None:
            return session.request(*args, **kwargs)
        return super(FuturesSessionFlex, self).request(*args, **kwargs)

    def close(self):
        for key, session in self.__sessions.items():
            session.close()
        self.__default_sessions.close()
